try:
	from lxml import etree as ElementTree
	LXML = True
except ImportError:
	from xml.etree import ElementTree
	LXML = False
try:
	from aniso8601 import parse_datetime
except ImportError:
	from dateutil.parser import parse as parse_datetime
from xml.dom import minidom

from . import SYSTEM_DTD

__all__ = [
	"ElementTree", "annotate_replay", "parse_datetime", "pretty_xml", "toxml"
]


def toxml(root, pretty):
	if LXML:
		kwargs = {
			"doctype": '<!DOCTYPE hsreplay SYSTEM "%s">' % (SYSTEM_DTD),
			"pretty_print": pretty,
			"xml_declaration": True,
			"encoding": "utf-8",
		}
		xml = ElementTree.tostring(root, **kwargs)
		return xml.decode("utf-8")

	if pretty:
		return pretty_xml(root)
	return ElementTree.tostring(root).decode("utf-8")


def pretty_xml(root):
	xml = ElementTree.tostring(root)
	ret = minidom.parseString(xml)

	imp = minidom.DOMImplementation()
	doctype = imp.createDocumentType(
		qualifiedName="hsreplay",
		publicId="",
		systemId=SYSTEM_DTD,
	)
	doc = imp.createDocument(None, "HSReplay", doctype)
	for element in list(ret.documentElement.childNodes):
		doc.documentElement.appendChild(element)
	for k, v in root.attrib.items():
		doc.documentElement.setAttribute(k, v)

	ret = doc.toprettyxml(indent="\t")
	return "\n".join(line for line in ret.split("\n") if line.strip())


def _to_string(tag):
	result = "<%s" % tag.tag
	if len(tag.attrib):
		result += " "
		result += " ".join("%s=%s" % item for item in tag.attrib.items())
	result += ">"
	return result


class ResolvedString(str):
	is_resolved = True


def _get_card_name(db, card_id):
	if hasattr(card_id, "is_resolved") and card_id.is_resolved:
		return card_id
	if card_id not in db:
		return "Unknown card %s" % card_id
	return db[card_id].name


def annotate_replay(infile, outfile):
	from hearthstone import cardxml
	from hearthstone.enums import (
		BlockType, GameTag, MetaDataType, Mulligan, PlayState, State, Step, Zone
	)
	db, _ = cardxml.load()
	entities = {}

	entity_ref_tags = {
		GameTag.LAST_AFFECTED_BY,
		GameTag.LAST_CARD_PLAYED,
		GameTag.PROPOSED_ATTACKER,
		GameTag.PROPOSED_DEFENDER,
		GameTag.WEAPON,
	}

	tree = ElementTree.parse(infile)
	root = tree.getroot()

	for tag in root.iter("FullEntity"):
		if "cardID" in tag.attrib:
			entities[tag.attrib["id"]] = tag.attrib["cardID"]

	for tag in root.iter("GameEntity"):
		entities[tag.attrib["id"]] = ResolvedString("GameEntity")

	for tag in root.iter("Player"):
		entities[tag.attrib["id"]] = ResolvedString(tag.attrib["name"])

	for tag in root.iter("ShowEntity"):
		if "cardID" in tag.attrib:
			entities[tag.attrib["entity"]] = tag.attrib["cardID"]
			tag.set("EntityName", _get_card_name(db, tag.attrib["cardID"]))

	for tag in root.iter("FullEntity"):
		if tag.attrib["id"] in entities:
			tag.set("EntityName", _get_card_name(db, entities[tag.attrib["id"]]))

	block_counter = 1
	for tag in root.iter("Block"):
		tag.set("block_sequence_num", str(block_counter))
		block_counter += 1
		if "entity" in tag.attrib and tag.attrib["entity"] in entities:
			tag.set("EntityCardID", entities[tag.attrib["entity"]])
			if tag.attrib["entity"] == "1":
				tag.set("EntityCardName", "GameEntity")
			elif tag.attrib["entity"] in ("2", "3"):
				tag.set("EntityCardName", entities[tag.attrib["entity"]])
			else:
				tag.set("EntityCardName", _get_card_name(db, entities[tag.attrib["entity"]]))

		if "target" in tag.attrib and tag.attrib["target"] in entities:
			tag.set("TargetName", _get_card_name(db, entities[tag.attrib["target"]]))

		if "triggerKeyword" in tag.attrib:
			try:
				tag.set("TriggerKeywordName", GameTag(int(tag.attrib["triggerKeyword"])).name)
			except ValueError:
				pass

	for tag in root.iter("Tag"):
		try:
			name = GameTag(int(tag.attrib["tag"])).name
			tag.set("GameTagName", name)

			if name == "STEP":
				tag.set("StepName", Step(int(tag.attrib["value"])).name)

			if name == "ZONE":
				tag.set("ZoneName", Zone(int(tag.attrib["value"])).name)

		except ValueError:
			pass

	for tag_change in root.iter("TagChange"):
		if "entity" in tag_change.attrib and tag_change.attrib["entity"] in entities:
			tag_change.set("EntityCardID", entities[tag_change.attrib["entity"]])
			tag_change.set("EntityCardName", _get_card_name(db, entities[tag_change.attrib["entity"]]))

		if int(tag_change.attrib["tag"]) in entity_ref_tags and tag_change.attrib["value"] in entities:
			tag_change.set("ValueReferenceCardID", entities[tag_change.attrib["value"]])
			tag_change.set(
				"ValueReferenceCardName",
				_get_card_name(db, entities[tag_change.attrib["value"]])
			)

		if tag_change.attrib["tag"] == str(GameTag.STATE.value):
			tag_change.set("StateName", State(int(tag_change.attrib["value"])).name)

		if tag_change.attrib["tag"] == str(GameTag.PLAYSTATE.value):
			tag_change.set("PlayStateName", PlayState(int(tag_change.attrib["value"])).name)

		try:
			name = GameTag(int(tag_change.attrib["tag"])).name
			tag_change.set("GameTagName", name)

			if name == "STEP":
				tag_change.set("StepName", Step(int(tag_change.attrib["value"])).name)

			if name == "NEXT_STEP":
				tag_change.set("StepName", Step(int(tag_change.attrib["value"])).name)

			if name == "ZONE":
				tag_change.set("ZoneName", Zone(int(tag_change.attrib["value"])).name)

			if name == "MULLIGAN_STATE":
				tag_change.set("MulliganStateName", Mulligan(int(tag_change.attrib["value"])).name)

		except ValueError:
			pass

	for block in root.iter("Block"):
		try:
			name = BlockType(int(block.attrib["type"])).name
		except ValueError:
			name = block.attrib["type"]

		block.set("BlockTypeName", name)

	for option in root.iter("Option"):
		if "entity" in option.attrib and option.attrib["entity"] in entities:
			option.set("EntityCardID", entities[option.attrib["entity"]])
			option.set("EntityName", _get_card_name(db, entities[option.attrib["entity"]]))

	for target in root.iter("Target"):
		if "entity" in target.attrib and target.attrib["entity"] in entities:
			target.set("EntityCardID", entities[target.attrib["entity"]])
			target.set("EntityName", _get_card_name(db, entities[target.attrib["entity"]]))

	for choices in root.iter("Choices"):
		if "entity" in choices.attrib and choices.attrib["entity"] in entities:
			choices.set("EntityCardID", entities[choices.attrib["entity"]])
			choices.set("EntityName", _get_card_name(db, entities[choices.attrib["entity"]]))
		if "source" in choices.attrib and choices.attrib["source"] in entities:
			choices.set("SourceCardID", entities[choices.attrib["source"]])
			choices.set("SourceName", _get_card_name(db, entities[choices.attrib["source"]]))

	for choice in root.iter("Choice"):
		if "entity" in choice.attrib and choice.attrib["entity"] in entities:
			choice.set("EntityCardID", entities[choice.attrib["entity"]])
			choice.set("EntityName", _get_card_name(db, entities[choice.attrib["entity"]]))

	for meta in root.iter("MetaData"):
		if "meta" in meta.attrib:
			meta.set("MetaName", MetaDataType(int(meta.attrib["meta"])).name)

	for target in root.iter("Info"):
		if "entity" in target.attrib and target.attrib["entity"] in entities:
			target.set("EntityName", _get_card_name(db, entities[target.attrib["entity"]]))

	tree.write(outfile, pretty_print=True)



import logging
from typing import Optional

from hearthstone.enums import MetaDataType
from hslog import LogParser
from hslog.packets import (Block, CachedTagForDormantChange, ChangeEntity,
                           Choices, ChosenEntities, CreateGame, FullEntity,
                           HideEntity, MetaData, Options, ResetGame,
                           SendChoices, SendOption, ShowEntity, ShuffleDeck,
                           SubSpell, TagChange, VOSpell)
from hslog.player import coerce_to_entity_id, PlayerManager

from . import elements


def serialize_entity(entity):
    if entity:
        return coerce_to_entity_id(entity)


def add_initial_tags(ts, packet, packet_element):
    for tag, value in packet.tags:
        tag_element = elements.TagNode(ts, tag, value)
        packet_element.append(tag_element)


def add_choices(ts, packet, packet_element):
    for i, entity_id in enumerate(packet.choices):
        choice_element = elements.ChoiceNode(ts, i, entity_id)
        packet_element.append(choice_element)


def add_options(ts, packet, packet_element):
    for i, option in enumerate(packet.options):
        if option.optype == "option":
            cls = elements.OptionNode
        elif option.optype == "target":
            cls = elements.OptionTargetNode
        elif option.optype == "subOption":
            cls = elements.SubOptionNode
        else:
            raise NotImplementedError("Unhandled option type: %r" % (option.optype))
        try:
            entity = serialize_entity(option.entity)
        except RuntimeError:
            # This is a hack to ensure we can serialize games from Hearthstone 18336.
            # Real names are shoved in the options, not used anywhere else...
            entity = None
        option_element = cls(ts, i, entity, option.error, option.error_param, option.type)
        add_options(ts, option, option_element)
        packet_element.append(option_element)


def add_packets_recursive(packets, entity_element):
    for packet in packets:
        if hasattr(packet, "entity"):
            _ent = serialize_entity(packet.entity)
        ts = packet.ts

        if isinstance(packet, CreateGame):
            packet_element = elements.GameEntityNode(ts, _ent)
            add_initial_tags(ts, packet, packet_element)
            entity_element.append(packet_element)
            for player in packet.players:
                entity_id = serialize_entity(player.entity)
                player_element = elements.PlayerNode(
                    ts, entity_id, player.player_id,
                    player.hi, player.lo, player.name
                )
                entity_element.append(player_element)
                add_initial_tags(ts, player, player_element)
            continue
        elif isinstance(packet, Block):
            effect_index = int(packet.effectindex or 0)
            packet_element = elements.BlockNode(
                ts, _ent, packet.type,
                packet.index if packet.index != -1 else None,
                packet.effectid or None,
                effect_index if effect_index != -1 else None,
                serialize_entity(packet.target),
                packet.suboption if packet.suboption != -1 else None,
                packet.trigger_keyword if packet.trigger_keyword else None
            )
            add_packets_recursive(packet.packets, packet_element)
        elif isinstance(packet, MetaData):
            # With verbose=false, we always have 0 packet.info :(
            if len(packet.info) not in (0, packet.count):
                logging.warning("META_DATA count is %r for %r", packet.count, packet.info)

            if packet.meta == MetaDataType.JOUST:
                data = serialize_entity(packet.data)
            else:
                data = packet.data
            packet_element = elements.MetaDataNode(
                ts, packet.meta, data, packet.count
            )
            for i, info in enumerate(packet.info):
                e = elements.MetaDataInfoNode(packet.ts, i, info)
                packet_element.append(e)
        elif isinstance(packet, TagChange):
            packet_element = elements.TagChangeNode(
                packet.ts, _ent, packet.tag, packet.value,
                packet.has_change_def if packet.has_change_def else None
            )
        elif isinstance(packet, HideEntity):
            packet_element = elements.HideEntityNode(ts, _ent, packet.zone)
        elif isinstance(packet, ShowEntity):
            packet_element = elements.ShowEntityNode(ts, _ent, packet.card_id)
            add_initial_tags(ts, packet, packet_element)
        elif isinstance(packet, FullEntity):
            packet_element = elements.FullEntityNode(ts, _ent, packet.card_id)
            add_initial_tags(ts, packet, packet_element)
        elif isinstance(packet, ChangeEntity):
            packet_element = elements.ChangeEntityNode(ts, _ent, packet.card_id)
            add_initial_tags(ts, packet, packet_element)
        elif isinstance(packet, Choices):
            packet_element = elements.ChoicesNode(
                ts, _ent, packet.id, packet.tasklist, packet.type,
                packet.min, packet.max, serialize_entity(packet.source)
            )
            add_choices(ts, packet, packet_element)
        elif isinstance(packet, SendChoices):
            packet_element = elements.SendChoicesNode(ts, packet.id, packet.type)
            add_choices(ts, packet, packet_element)
        elif isinstance(packet, ChosenEntities):
            packet_element = elements.ChosenEntitiesNode(ts, _ent, packet.id)
            add_choices(ts, packet, packet_element)
        elif isinstance(packet, Options):
            packet_element = elements.OptionsNode(ts, packet.id)
            add_options(ts, packet, packet_element)
        elif isinstance(packet, SendOption):
            packet_element = elements.SendOptionNode(
                ts, packet.option, packet.suboption, packet.target, packet.position
            )
        elif isinstance(packet, ResetGame):
            packet_element = elements.ResetGameNode(ts)
        elif isinstance(packet, SubSpell):
            packet_element = elements.SubSpellNode(
                ts, packet.spell_prefab_guid, packet.source, packet.target_count
            )
            for i, target in enumerate(packet.targets):
                e = elements.SubSpellTargetNode(ts, i, target)
                packet_element.append(e)
            add_packets_recursive(packet.packets, packet_element)
        elif isinstance(packet, CachedTagForDormantChange):
            packet_element = elements.CachedTagForDormantChangeNode(
                packet.ts, _ent, packet.tag, packet.value
            )
        elif isinstance(packet, VOSpell):
            packet_element = elements.VOSpellNode(
                packet.ts, packet.brguid, packet.vospguid, packet.blocking, packet.delayms
            )
        elif isinstance(packet, ShuffleDeck):
            packet_element = elements.ShuffleDeckNode(packet.ts, packet.player_id)
        else:
            raise NotImplementedError(repr(packet))
        entity_element.append(packet_element)


def parse_log(fp, processor, date):
    parser = LogParser()
    parser._game_state_processor = processor
    parser._current_date = date
    parser.read(fp)

    return parser


def game_to_xml(
    tree,
    game_meta=None,
    player_manager: Optional[PlayerManager] = None,
    player_meta=None,
    decks=None
):
    # game_tree = tree.export()
    game_element = elements.GameNode(tree.ts)
    add_packets_recursive(tree.packets, game_element)
    players = game_element.players

    if game_meta is not None:
        game_element._attributes = game_meta

    if player_meta is not None:
        for player, meta in zip(players, player_meta):
            player._attributes = meta

    if decks is not None:
        for player, deck in zip(players, decks):
            player.deck = deck

        if player_manager:
            # Set the player names
            for player in players:
                if not player.name:
                    player.name = player_manager.get_player_by_entity_id(player.id).name

    return game_element



from . import DTD_VERSION
from .dumper import game_to_xml, parse_log
from .elements import GameNode
from .utils import ElementTree, toxml


class HSReplayDocument:
	ROOT_NAME = "HSReplay"

	@classmethod
	def from_log_file(cls, fp, processor="GameState", date=None, build=None):
		parser = parse_log(fp, processor, date)
		return cls.from_parser(parser, build)

	@classmethod
	def from_packet_tree(cls, packet_tree, build=None, player_manager=None):
		ret = cls(build)
		ret._update_document()
		for tree in packet_tree:
			game = game_to_xml(tree, player_manager=player_manager)
			ret.games.append(game)
		return ret

	@classmethod
	def from_parser(cls, parser, build=None):
		return cls.from_packet_tree(
			parser.games,
			build,
			player_manager=parser.player_manager
		)

	@classmethod
	def from_xml_file(cls, fp):
		xml = ElementTree.parse(fp)
		return cls.from_xml(xml)

	@classmethod
	def from_xml(cls, xml):
		root = xml.getroot()
		build = root.attrib.get("build")
		ret = cls(build)
		ret.version = root.attrib.get("version")
		for game in xml.findall("Game"):
			gamenode = GameNode.from_xml(game)
			ret.games.append(gamenode)

		return ret

	def __init__(self, build=None):
		self.build = build
		self.version = DTD_VERSION
		self.games = []
		self.root = None

	def _create_document(self):
		builder = ElementTree.TreeBuilder()
		attrs = {"version": self.version}
		if self.build is not None:
			attrs["build"] = str(self.build)
		builder.start(self.ROOT_NAME, attrs)
		builder.end(self.ROOT_NAME)

		self.root = builder.close()
		return self.root

	def _update_document(self):
		self._create_document()
		for game in self.games:
			self.root.append(game.xml())

	def to_packet_tree(self):
		self._update_document()
		ret = []
		for game in self.games:
			ret.append(game.export())
		return ret

	def to_xml(self, pretty=False):
		self._update_document()
		return toxml(self.root, pretty=pretty)
